@@ -2,11 +2,64 @@
 
 BLOCK_SIZE = 1048576 * 4
 
-import os, sys, hashlib, lzma, aes, time, zlib, bz2, timeit
-import cPickle as pickle
-import adler64
-import multiprocessing
+import os, sys, hashlib, time, bz2, base64
+try: import cPickle as pickle
+except: import pickle
 from itertools import ifilter
+from nimbstor.optimized import rolling_search, stor_buffer
+
+compress_function_bz2_method = lambda b: bz2.compress(b, 9)
+compress_function_methods = {}
+def compress_function(compression, meta):
+  if compression: compression = tuple(compression)
+  else: compression = (None, 0)
+  if meta: return compress_function_bz2_method
+  elif compress_function_methods.has_key(compression):
+    return compress_function_methods[compression]
+  elif compression == ('lzma', 10):
+    import lzma
+    compress_function_methods[compression] = lambda b: lzma.compress(b, {'level': 9, 'extreme': True})
+  elif compression[0] == 'lzma':
+    import lzma
+    compress_function_methods[compression] = lambda b: lzma.compress(b, {'level': compression[1]})
+  elif compression[0] == 'gzip':
+    import zlib
+    compress_function_methods[compression] = lambda b: zlib.compress(b, compression[1])
+  elif compression[0] == 'bzip2':
+    compress_function_methods[compression] = lambda b: bz2.compress(b, compression[1])
+  elif compression[0] == None:
+    compress_function_methods[compression] = lambda b: b
+  else: raise SystemError("Unsupported compression type %s" % repr(compression))
+  return compress_function_methods[compression]
+
+decompress_function_bz2_method = lambda b: bz2.decompress(b)
+decompress_function_methods = {}
+def decompress_function(compression, meta):
+  if compression: compression = tuple(compression)
+  else: compression = (None, 0)
+  if meta: return decompress_function_bz2_method
+  elif decompress_function_methods.has_key(compression[0]):
+    return decompress_function_methods[compression[0]]
+  elif compression[0] == 'lzma':
+    import lzma
+    decompress_function_methods[compression[0]] = lambda b: lzma.decompress(b)
+  elif compression[0] == 'gzip':
+    import zlib
+    decompress_function_methods[compression[0]] = lambda b: zlib.decompress(b)
+  elif compression[0] == 'bzip2':
+    decompress_function_methods[compression[0]] = lambda b: bz2.decompress(b)
+  elif compression[0] == None:
+    decompress_function_methods[compression[0]] = lambda b: b
+  else: raise SystemError("Unsupported compression type %s" % repr(compression))
+  return decompress_function_methods[compression[0]]
+
+aes_keysetup_aes = None
+def aes_keysetup(password):
+  global aes_keysetup_aes
+  if aes_keysetup_aes == None:
+    import aes
+    aes_keysetup_aes = aes.Keysetup
+  return aes_keysetup_aes(password)
 
 def message_encode(buffer, password, meta, compression, bufnum, checksum1, checksum2):
   """Internal function to encode message.
@@ -17,19 +70,13 @@ def message_encode(buffer, password, meta, compression, bufnum, checksum1, check
     meta (boolean): Meta blocks are handled somewhat different.
     compression (tuple): Compression type to use.
     bufnum (int): Current buffer number
-    checksum1 (int): Adler64 checksum.
+    checksum1 (int): Rolling checksum.
     checksum2 (str): Checksum as returned by stor_stream.checksum(buffer)
   """
-  if meta: buffer = bz2.compress(buffer, 9)
-  elif compression == ('lzma', 10): buffer = lzma.compress(buffer, {'level': 9, 'extreme': True})
-  elif compression[0] == 'lzma': buffer = lzma.compress(buffer, {'level': compression[1]})
-  elif compression[0] == 'gzip': buffer = zlib.compress(buffer, compression[1])
-  elif compression[0] == 'bzip2': bffer = bz2.compress(buffer, compression[1])
-  elif compression[0] != None:
-    raise SystemError("Unsupported compression type %s" % repr(compression))
+  buffer = apply(compress_function(compression, meta), [buffer])
   bufrest = len(buffer) % 16
   if bufrest > 0: buffer = buffer + "\0" * (16 - bufrest)
-  if password: buffer = aes.Keysetup(password).encrypt(buffer)
+  if password: buffer = aes_keysetup(password).encrypt(buffer)
   return (buffer, bufnum, checksum1, checksum2)
 
 def message_decode(buffer, password, cksumpwd, meta, compression, bufnum, checksum1, checksum2, bufflen):
@@ -42,175 +89,27 @@ def message_decode(buffer, password, cksumpwd, meta, compression, bufnum, checks
     meta (boolean): Meta blocks are handled somewhat different.
     compression (tuple): Compression type to use.
     bufnum (int): Current buffer number
-    checksum1 (int): Adler64 checksum.
+    checksum1 (int): Rolling checksum.
     checksum2 (str): Checksum as returned by stor_stream.checksum(buffer)
     bufflen (str): Correct data length to return or 0.
   """
   try:
-    if password: buffer = aes.Keysetup(password).decrypt(buffer)
-    if meta: buffer = bz2.decompress(buffer)
-    elif compression[0] == 'lzma': buffer = lzma.decompress(buffer)
-    elif compression[0] == 'gzip': buffer = zlib.decompress(buffer)
-    elif compression[0] == 'bzip2': buffer = bz2.decompress(buffer)
-    elif compression[0] != None:
-      raise SystemError("Unsupported compression type %s" % repr(compression))
+    if password: buffer = aes_keysetup(password).decrypt(buffer)
+    buffer = apply(decompress_function(compression, meta), [buffer])
     if bufflen != 0 and len(buffer) > bufflen: buffer = buffer[:bufflen]
     ctx = hashlib.sha256()
     if cksumpwd != None: ctx.update(cksumpwd)
     ctx.update(buffer)
-    newcksum = ctx.hexdigest().upper()
+    newcksum = base64.urlsafe_b64encode(ctx.digest())[:-1].upper()
     if newcksum != checksum2: raise SystemError()
   except Exception, err:
     raise SystemError("Wrong password or invalid data (%s)" % repr(err))
-  return (buffer, bufnum, checksum1, checksum2, bufflen)
-
-
-class stor_buffer:
-  """Manage buffer for nimbstor.
-
-  Not a performant implementation, should be optimized one day.
-
-  """
-
-  def __init__(self, data = "", splitsize = None):
-    """Initialize the buffer.
-
-    Args:
-      data (str): Preset buffer content.
-
-    >>> stor_buffer("asdf").value()
-    'asdf'
-    """
-    self._data = [data]
-    self._len = len(data)
-    self._val = None
-    self._splitsize = splitsize
-
-  def __len__(self):
-    """Return buffer content length.
-
-    >>> len(stor_buffer("asdfqw"))
-    6
-    """
-    return self._len
-
-  def push(self, data, left = False):
-    """Push data to the buffer.
-
-    Args:
-      data (str): Data to be pushed.
-      left (bool): Push to begin of the buffer (default: False)
-
-    >>> buf = stor_buffer('asdf')
-    >>> buf.push('qwer')
-    >>> buf.push('zxcv', True)
-    >>> buf.value()
-    'zxcvasdfqwer'
-    """
-    self._len += len(data)
-    if self._splitsize != None:
-      listdat = []
-      curpos = 0
-      while curpos < self._len:
-        listdat.append(data[curpos:curpos + self._splitsize])
-        curpos += self._splitsize
-      if left: self._data[0:0] = listdat
-      else: self._data.extend(listdat)
-    else:
-      if left: self._data.insert(0, data)
-      else: self._data.append(data)
-    self._val = None
-
-  def _rmsize(self, size):
-    """Internal function to remove some data from buffer.
-
-    Args:
-      size (int): Amount of data to remove.
-
-    This function expects, that self._val has correct value.
-
-    >>> buf = stor_buffer('asdfqwer')
-    >>> buf.value()
-    'asdfqwer'
-    >>> buf._rmsize(3)
-    >>> buf.value()
-    'fqwer'
-    """
-    new = self._val[size:]
-    self._val = None
-    self._data = [new]
-    self._len = len(new)
-
-  def pop(self, size = None):
-    """Pop some data from buffer.
-
-    Args:
-      size (int): Amount of data to pop.
-
-    >>> buf = stor_buffer('asdfqwer')
-    >>> buf.pop(3)
-    'asd'
-    >>> buf.value()
-    'fqwer'
-    """
-    if size == None:
-      buf = "".join(self._data)
-      self._data = []
-      self._len = 0
-      self._val = None
-      return buf
-    if len(self._data) == 0:
-      return ''
-    self._val = None
-    buf = []
-    restsize = size
-    while restsize > 0 and len(self._data) > 0:
-      curbuf = self._data[0]
-      curlen = len(curbuf)
-      if curlen > restsize:
-        buf.append(curbuf[:restsize])
-        self._data[0] = curbuf[restsize:]
-        self._len -= restsize
-        restsize = 0
-      else:
-        buf.append(curbuf)
-        del self._data[0]
-        self._len -= curlen
-        restsize -= curlen
-    return ''.join(buf)
-
-  def truncate(self, size = None):
-    """Remove some amount of data from buffer.
-
-    Args:
-      size (int): Amount of data to remove (default: all).
-
-    >>> buf = stor_buffer('asdfqwer')
-    >>> buf.truncate(3)
-    >>> buf.value()
-    'fqwer'
-    """
-    if size == None:
-      self._data = []
-      self._val = None
-      self._len = 0
-      return
-    if self._val == None: self._val = "".join(self._data)
-    self._rmsize(size)
-
-  def value(self):
-    """Return the buffer value.
-
-    >>> stor_buffer('asdf').value()
-    'asdf'
-    """
-    if self._val == None: self._val = "".join(self._data)
-    return self._val
+  return buffer
 
 class stor_stream:
   """Base class for storage streams."""
 
-  def __init__(self, password = None):
+  def __init__(self, password = None, verbose = False):
     """Initialize base stream.
 
     Args:
@@ -219,6 +118,7 @@ class stor_stream:
     >>> stor_stream('asdf')._pass == stor_stream('qwer')._pass
     False
     """
+    self._closed = False
     self._pos = 0
     self._known_blocks = {}
     self._known_parts = {}
@@ -231,6 +131,10 @@ class stor_stream:
       self._aes = None
       self._pass = None
     self._archive = None
+    self._verbose = verbose
+
+  def verbose(self):
+    return self._verbose
 
   def open(self, backend):
     """Initialize and open the stream.
@@ -259,7 +163,7 @@ class stor_stream:
     ctx = hashlib.sha256()
     if self._pass != None: ctx.update(self._pass)
     ctx.update(data)
-    return ctx.hexdigest().upper()
+    return base64.urlsafe_b64encode(ctx.digest())[:-1].upper()
 
   def tell(self):
     """Return the current position in stream."""
@@ -278,28 +182,41 @@ class stor_stream:
 
     Args:
       number (int): Block number.
-      checksum1 (int): Adler64 checksum.
+      checksum1 (int): Rolling checksum.
       checksum2 (str): Block checksum as returned by self.checksum(data)
       backend (object): Backend reference, which can read this block.
 
     Backends use this function to register known blocks.
     """
     checksum1 = int(checksum1, 16)
-    if number == 0 and checksum2 not in self._known_metadata: self._known_metadata.append(checksum2)
-    elif checksum1 == 0: self._known_parts.setdefault(checksum2, []).append(number)
-    else: self._known_blocks.setdefault(checksum1, {}).setdefault(checksum2, []).append(number)
+    if number == 0 and checksum2 not in self._known_metadata:
+      self._known_metadata.append(checksum2)
+    elif checksum1 == 0: self._known_parts[checksum2] = number
+    else: self._known_blocks.setdefault(checksum1, {})[checksum2] = number
 
-  def close(self):
+  def close(self, dont_commit = False):
     """Close the stream.
 
     The backends close will be also called.
     """
-    self._backend.close()
+    if self._closed: return
+    if self._backend != None:
+      self._backend.close(dont_commit)
+
+  def blocks(self):
+    return [(0, "%08X" % 0, self._archive_checksum, 0)] + self._metadata['blocks'][:]
+
+  def description(self): return self._metadata['description']
+  def keywords(self): return self._metadata['keywords']
+  def parent(self): return self._metadata['parent']
+  def compression(self): return self._metadata['compression']
+  def size(self): return self._metadata['size']
+  def usage(self): return self._metadata['usage']
 
 class stor_output(stor_stream):
   """Storage output stream."""
 
-  def __init__(self, password, compression, parent, description, keywords, parlimit = 4, onlyifnew = False):
+  def __init__(self, password, compression, parent, description, keywords, parlimit = 1, onlyifnew = False, verbose = False):
     """Initialization of output stream.
 
     Args:
@@ -310,7 +227,7 @@ class stor_output(stor_stream):
       keywords (list): List of keyword strings.
       onlyifnew (boolean): Set to True to commit empty (fully deduplicated) archives.
     """
-    stor_stream.__init__(self, password)
+    stor_stream.__init__(self, password, verbose)
     self._metadata = {
 	'description': description,
 	'keywords': keywords,
@@ -321,10 +238,12 @@ class stor_output(stor_stream):
 	'blocks': [],
 	'size': 0,
     }
-    self._procslimit = parlimit
-    if parlimit != None:
-      self._procspool = multiprocessing.Pool()
+    if parlimit > 1:
+      self._procslimit = parlimit
+      import multiprocessing
+      self._procspool = multiprocessing.Pool(parlimit)
       self._procslist = []
+    else: self._procslimit = None
     self._buffer = stor_buffer()
     self._bufnum = 1
     self._bufsize = BLOCK_SIZE * 2
@@ -344,10 +263,6 @@ class stor_output(stor_stream):
     self._buffer.push(buffer)
     self.flush_buffer(self._bufsize)
 
-  def size(self):
-    """Returns the size (length) of stream."""
-    return self._metadata['usage']
-
   def set_metainfo(self, info):
     """Set the meta information, user defined."""
     self._metadata['metainfo'] = info
@@ -356,78 +271,88 @@ class stor_output(stor_stream):
     """Set the parent archive name."""
     self._metadata['parent'] = parent
 
-  def _block_already_exists(self, buffer, begin, end, checksum1):
+  def _block_already_exists(self, data, pos, checksum1):
     """Checks if the given block already exists.
 
     Args:
-      buffer (str): Data buffer.
-      begin (int): Block start position.
-      end (int): Block end position.
+      data (str): Data buffer.
+      pos (int): Block start position.
+      checksum1 (int): Rolling checksum of the data.
 
-    Call only if the block with given Adler64 checksum really exist (it means
+    Call only if the block with given rolling checksum really exist (it means
     that it was registred with append_block_info).
     """
-    checksum2 = ctx.hexdigest().upper()
+    checksum2 = self.checksum(data)
+    #print "@ block found: %08X (%s)" % (checksum1, checksum2)
     if self._known_blocks[checksum1].has_key(checksum2):
       self._checksum2 = checksum2
       self._checksumblock = self._known_blocks[checksum1][checksum2]
+      if self._verbose:
+        print "@ deduplication:", checksum2
       return True
     return False
 
-  def close(self):
+  def close(self, dont_commit = False):
     """Close the output stream.
 
     This function write also the metadata, without metadata the archive does
     not exist.
     """
-    self.flush_buffer(BLOCK_SIZE)
-    self.commit_buffer(self._buffer.pop(), 0)
-    if not self._onlyifnew or self._something_written:
-      self._finish_write(0)
-      self._metadata['usage'] = self._size
-      self._metadata['size'] = self._pos
-      buffer = pickle.dumps(self._metadata)
-      self._bufnum = 0
-      self.commit_buffer(buffer, 0, meta = True)
-      self._finish_write(0)
-    stor_stream.close(self)
+    if self._closed: return
+    try:
+      try:
+        if self._backend != None and dont_commit != True:
+          self.flush_buffer(BLOCK_SIZE)
+          self.commit_buffer(self._buffer.pop(), 0)
+          if not self._onlyifnew or self._something_written:
+            self._finish_write(0)
+            self._metadata['usage'] = self._size
+            self._metadata['size'] = self._pos
+            buffer = pickle.dumps(self._metadata)
+            self._bufnum = 0
+            self.commit_buffer(buffer, 0, meta = True)
+            self._finish_write(0)
+      finally: stor_stream.close(self, dont_commit)
+    finally:
+      if self._procslimit != None:
+        self._procspool.terminate()
+        self._procspool.join()
 
   def flush_buffer(self, tosize):
     """Flush data from buffer.
     
-    This function do the deduplication with Adler64.
+    This function do the deduplication with rolling checksum.
     """
     while len(self._buffer) >= tosize:
       buffer = self._buffer.value()
-      found, pos, dest, fcks, cks = adler64.search(self._known_blocks, self._block_already_exists, buffer, BLOCK_SIZE)
-      if found:
-	destlen = len(dest)
-	if destlen == BLOCK_SIZE: self.commit_buffer(dest, fcks)
-	elif destlen > 0: self.commit_buffer(dest, 0)
-	checksum1 = "%016X" % cks
-	self._metadata['blocks'].append((self._checksumblock, checksum1, self._checksum2, BLOCK_SIZE))
-	self._bufnum += 1
-	self._checksum2 = None
-	self._buffer.truncate(pos + BLOCK_SIZE + destlen)
-      else:
-	self.commit_buffer(self._buffer.pop(BLOCK_SIZE), fcks)
+      pos, cks = rolling_search(self._known_blocks, self._block_already_exists, buffer, BLOCK_SIZE)
+      if pos != None:
+        if pos > 0: self.commit_buffer(buffer[:pos], 0)
+        data = buffer[pos:tosize]
+        self._metadata['blocks'].append((self._checksumblock, "%08X" % cks, self._checksum2, BLOCK_SIZE))
+        self._bufnum += 1
+        self._found_buffer = None
+        self._checksum2 = None
+        self._buffer.truncate(pos + BLOCK_SIZE)
+      else: self.commit_buffer(self._buffer.pop(BLOCK_SIZE), cks)
 
   def _finish_write(self, limit):
-    while len(self._procslist) > limit:
-      proc = self._procslist.pop(0)
-      buffer, bufnum, checksum1, checksum2 = proc.get()
-      self._size += len(buffer)
-      self._backend.write_buffer(buffer, bufnum, checksum1, checksum2)
+    if self._procslimit != None:
+      while len(self._procslist) > limit:
+        proc = self._procslist.pop(0)
+        buffer, bufnum, checksum1, checksum2 = proc.get()
+        self._size += len(buffer)
+        self._backend.write_buffer(buffer, bufnum, checksum1, checksum2)
 
   def commit_buffer(self, buffer, checksum1int, meta = False):
     """Write data to the beckend.
 
     Args:
       buffer (str): Block buffer.
-      checksum1int (int): Adler64 checksum of the buffer.
+      checksum1int (int): Rolling checksum of the buffer.
       meta (boolean): Set to True if the buffer contains metadata.
     """
-    checksum1 = "%016X" % checksum1int
+    checksum1 = "%08X" % checksum1int
     checksum2 = self.checksum(buffer)
     buflen = len(buffer)
     if checksum1int == 0 and self._known_parts.has_key(checksum2):
@@ -457,18 +382,20 @@ class stor_output(stor_stream):
 class stor_input(stor_stream):
   """Storage input stream."""
 
-  def __init__(self, password, archive, parlimit = 4):
+  def __init__(self, password, archive, parlimit = 1, verbose = False):
     """Initialization of input stream.
 
     Args:
       password (str): Password for decrypting data (or None).
       archive (str): Name of the archive (checksum).
     """
-    stor_stream.__init__(self, password)
-    self._procslimit = parlimit
-    if parlimit != None:
+    stor_stream.__init__(self, password, verbose)
+    if parlimit > 1:
+      import multiprocessing
       self._procspool = multiprocessing.Pool()
       self._procslist = []
+      self._procslimit = parlimit
+    else: self._procslimit = None
     self._archive_checksum = archive
     self._buffer = stor_buffer(splitsize = 10240)
     self._archive = archive
@@ -483,26 +410,23 @@ class stor_input(stor_stream):
     stor_stream.open(self, backend)
     if self._archive_checksum not in self._known_metadata:
       raise SystemError("Archive %s not found." % repr(self._archive_checksum))
-    ck1 = "%016X" % 0
+    ck1 = "%08X" % 0
     ck2 = self._archive_checksum
-    bufgen = self.read_buffer(0, ck1, ck2)
-    self._metadata = None
-    for buffer in bufgen:
-      try:
-        buffer, _, _, _, _ = message_decode(buffer, self._aes, self._pass, True, None, 0, ck1, ck2, 0)
-        self._metadata = pickle.loads(buffer)
-      except Exception, err: continue
-      break
-    if self._metadata == None:
-      raise SystemError('Failed to read metdata')
+    try:
+      data = self.read_buffer(0, ck1, ck2)
+      data = message_decode(data, self._aes, self._pass, True, None, 0, ck1, ck2, 0)
+      self._metadata = pickle.loads(data)
+    except: raise SystemError('Failed to read metdata')
     self._compression = self._metadata['compression']
     self._blocks = self._metadata['blocks'][:]
 
-  def bocks(self):
-    return [(0, "%016X" % 0, self._archive_checksum),] + sef._metadata['blocks']
-
-  def close(self):
-    stor_stream.close(self)
+  def close(self, dont_commit = False):
+    if self._closed: return
+    try: stor_stream.close(self, dont_commit)
+    finally:
+      if self._procslimit != None:
+        self._procspool.terminate()
+        self._procspool.join()
 
   def read(self, size):
     """Read data of given size.
@@ -512,43 +436,22 @@ class stor_input(stor_stream):
     data = None
     if self._procslimit != None:
       while len(self._buffer) < size and (len(self._blocks) > 0 or len(self._procslist) > 0):
-        data = None
         while len(self._procslist) < self._procslimit and len(self._blocks) > 0:
           num, ck1, ck2, blen = self._blocks.pop(0)
-          buffer_generator = self.read_buffer(num, ck1, ck2)
-          self._procslist.append((
-            buffer_generator,
+          data = self.read_buffer(num, ck1, ck2)
+          self._procslist.append(
             self._procspool.apply_async(message_decode, [
-              buffer_generator.next(),
-              self._aes, self._pass, False, self._compression,
-              num, ck1, ck2, blen,
-            ]),
-            num, ck1, ck2, blen))
-        if self._procslimit != None:
-          bufgen, proc, num, ck1, ck2, blen = self._procslist.pop(0)
-          try: data, num, ck1, ck2, blen = proc.get()
-          except Exception, err: data = None
-          while data == None:
-            data = bufgen.next()
-            try:
-              data, num, ck1, ck2, blen = apply(message_decode, [
-                data, self._aes, self._pass, False, self._compression,
-                num, ck1, ck2, blen,
-              ])
-            except Exception, err: data = None
+              data, self._aes, self._pass, False,
+              self._compression, num, ck1, ck2, blen]))
+        data = self._procslist.pop(0).get()
         self._buffer.push(data)
     else:
       while len(self._buffer) < size and len(self._blocks) > 0:
-        data = None
         num, ck1, ck2, blen = self._blocks.pop(0)
-        for buffer in self.read_buffer(num, ck1, ck2):
-          try:
-            data, num, ck1, ck2, blen = apply(message_decode, [
-              data, self._aes, self._pass, False, self._compression,
-              num, ck1, ck2, blen,
-            ])
-          except Exception, err: continue
-          break
+        data = self.read_buffer(num, ck1, ck2)
+        data = apply(message_decode, [
+              data, self._aes, self._pass, False,
+              self._compression, num, ck1, ck2, blen])
         self._buffer.push(data)
     x = time.time()
     data = self._buffer.pop(size)
@@ -566,24 +469,21 @@ class stor_util(stor_stream):
   remove, ...
   """
 
-  def __init__(self, password):
-    stor_stream.__init__(self, password)
+  def __init__(self, password, verbose = False):
+    stor_stream.__init__(self, password, verbose)
     self._errors = []
+    self._metadata = None
     self._metadata_data = {}
 
   def open(self, backend):
     stor_stream.open(self, backend)
-    checksum1 = "%016X" % 0
+    checksum1 = "%08X" % 0
     for metadata in self._known_metadata:
-      bufgen = self.read_buffer(0, checksum1, metadata)
       data = None
       try:
-        for buffer in bufgen:
-          try:
-            data, _, _, _, _ = message_decode(buffer, self._aes, self._pass, True, None, 0, checksum1, metadata, 0)
-            data = pickle.loads(data)
-          except Exception, err: continue
-          break
+        data = self.read_buffer(0, checksum1, metadata)
+        data = message_decode(data, self._aes, self._pass, True, None, 0, checksum1, metadata, 0)
+        data = pickle.loads(data)
       except Exception, err:
         self._errors.append((metadata, err))
         continue
@@ -607,7 +507,9 @@ class stor_util(stor_stream):
         weight = 0
         for kwd in keywords:
           for fnd in ifilter(lambda x: x == kwd, data['keywords']): weight += 10
-          for fnd in ifilter(lambda x: kwd in x and x != kwd, data['keywords']): weight += 2
+          for fnd in ifilter(lambda x: kwd in x and x != kwd, data['keywords']): weight += 3
+          if kwd in time.strftime("%F %R", time.localtime(data['timestamp'])): weight += 2
+          if kwd in str(data['size']): weight += 1
           if kwd.lower() in data['description'].lower(): weight += 5
           if kwd in repr(data['metainfo']): weight += 1
         result.append({
@@ -636,7 +538,7 @@ class stor_util(stor_stream):
     else: result.sort(key = lambda x: x['description'])
     return (result, parents, self._errors)
 
-  def keywords(self):
+  def keywordlist(self):
     """Return list of used keywords."""
     keywords = {}
     for metadata, data in self._metadata_data.items():
@@ -652,18 +554,55 @@ class stor_util(stor_stream):
         return True
     elif checksum1 == 0:
       if self._known_parts.has_key(checksum2):
-        if number in self._known_parts[checksum2]:
+        if number == self._known_parts[checksum2]:
           return True
     else:
       if self._known_blocks.has_key(checksum1):
         if self._known_blocks[checksum1].has_key(checksum2):
-          if number in self._known_blocks[checksum1][checksum2]:
+          if number == self._known_blocks[checksum1][checksum2]:
             return True
     return False
 
+  def description(self, archive):
+    return self._metadata_data[archive]['description']
+
+  def keywords(self, archive):
+    return self._metadata_data[archive]['keywords']
+
+  def parent(self, archive):
+    return self._metadata_data[archive]['parent']
+
+  def compression(self, archive):
+    return self._metadata_data[archive]['compression']
+
+  def size(self, archive):
+    return self._metadata_data[archive]['size']
+
+  def usage(self, archive):
+    return self._metadata_data[archive]['usage']
+
   def blocks(self, archive):
-    ck1 = "%016X" % 0
-    metadata = None
-    return [(0, ck1, archive, 0)] + self._metadata_data[archive]['blocks'][:]
+    self._metadata = self._metadata_data[archive]
+    self._archive_checksum = archive
+    try: ret = stor_stream.blocks(self)
+    finally:
+      self._metadata = None
+      self._archive_checksum = None
+    return ret
+
+  def verify_blocks(self, archive):
+    self._metadata = self._metadata_data[archive]
+    self._archive_checksum = archive
+    if self._verbose:
+      print "VERIFY ARCHIVE:", archive
+    try:
+      for num, cks1, cks2, size in stor_stream.blocks(self):
+        self.read_buffer(num, cks1, cks2)
+        if self._verbose:
+          print "OK %08d.%s.%s" % (num, cks1, cks2)
+    finally:
+      self._metadata = None
+      self._archive_checksum = None
 
 __all__ = ('stor_output', 'stor_input', 'stor_util')
+# vim:sts=2:sw=2:
